#!/usr/local/bin/python

import os, string, re, time, sys
from stat import *

cache_dir = '/tmp/psp-'

redo_all = 0
inside_html_block = 0
block_indent = 0


save_output = 0
lines = []

def emit_html (str):
    if save_output:
        lines.append (str)
    print str

def cache_exit():
    global cache_filename
    f = open (cache_filename, "w")
    for i in lines:
        f.write (i + '\n')
    f.close()

def cache_start (fn):
    global save_output
    global cache_filename
    save_output = 1
    cache_filename = fn
    sys.exitfunc = cache_exit

def cache_directive (key, timestamp):
    cache_fn = cache_dir + key.replace ('/', '_')
    # check if cache exists
    if not os.access (cache_fn, os.F_OK):
        # Nope. We need to build a new cache for this output
        cache_start (cache_fn)
        return
    # Check if cache is newer than timestamp
    cache_timestamp = os.stat (cache_fn)[ST_MTIME]
    if timestamp == None or cache_timestamp < timestamp:
        # Nope. Cache is outdated. Need to capture the output
        cache_start (cache_fn)
        return
    # We have a cache and it is up-to-date
    f = open (cache_fn)
    line = f.readline()
    while line:
        print line
        line = f.readline()
    f.close()
    # OK. Now we can exit right away
    sys.exit()
    
def check_for_cache_directive (line):
    left = line.find ('<%@')
    if left == -1:
        return ""
    right = line.find ('%>')
    middle = line[left + 3:right]
    return 'psp.cache_directive (' + middle + ')'

def replace_args (line):
    line = string.replace (line, '"', '\\"')
    
    left = line.find ('<%')

    while left != -1:
        right = line.find ('%>')
        middle = line[left+2:right]
        if middle[0] == '@':
            sub = 'psp.cache_directive (' + middle[1:] + ')'
        else:
            sub = 'str (' + middle + ')'
        line = line[0:left] + '" + ' + sub + ' + "' + line[right+2:]
        left = line.find ('<%')

    return line


def process_one_line (line):
    global inside_html_block
    global block_indent

    pattern_open_html = re.compile ('[ ]*<![ ]*')
    pattern_close_html = re.compile ('[ ]*!>[ ]*')

    if pattern_open_html.match (line):
        inside_html_block = 1
        block_indent = line.find ('<')
        return ""
    
    if pattern_close_html.match (line):
        inside_html_block = 0
        return ""

    if inside_html_block == 0:
        return line

    cache_dir = check_for_cache_directive (line)
    if cache_dir != "":
        return block_indent * ' ' + cache_dir
    
    line = replace_args (line)
    if line[0:block_indent] == block_indent * ' ':
        line = line[block_indent:]

    return block_indent * ' ' + 'psp.emit_html ("' + line + '")'
    

def file_needs_updating (in_file, out_file):
    if redo_all:
        return 1
    
    in_file_mod = os.stat (in_file)[ST_MTIME]

    try:
        # if file doesn't exist, exception is thrown
        out_file_mod = os.stat (out_file)[ST_MTIME]
    except:
        return 1

    return in_file_mod > out_file_mod


def write_helper_function (out_file):
    out_file.write ('import psp\n')
    #out_file.write ('\nimport process_html\n\n')
    #out_file.write ('def gen_html (str):\n')
    #out_file.write ('    print process_html.process_one_line (str)\n\n')

                    
def process_file (in_file, out_file):
    if not (file_needs_updating (in_file, out_file)):
        return
    
    print "Processing:", in_file
    inf = open (in_file)
    outf = open (out_file, "w")

    line = inf.readline()
    first_line = 1
    
    while line:
        line = re.sub ('\n', '', line)
        processed_line = process_one_line (line)

        if first_line and line[:2] != '#!':
            write_helper_function (outf)
        
        outf.write (processed_line)
        outf.write ('\n')

        if first_line and line[:2] == '#!':
            write_helper_function (outf)

        first_line = 0
        line = inf.readline()

    outf.close()

    
def visit (file):
    i = string.rfind (file, ".psp")
    if i == len (file) - 4:
        process_file (file, file[0:i] + '.py')

    
def traverse (path):
    files = os.listdir (path)
    for i in files:
        fn = path + "/" + i
        if os.path.isdir (fn):
            traverse (fn)
        else:
            visit (fn)

if __name__ == '__main__':
    if len(sys.argv) == 2:
        if sys.argv[1] != "--all":
            sys.stderr.write ('Usage: ' + sys.argv[0] + ' [--all]\n')
            sys.exit(1)
        redo_all = 1

    traverse (".")
