#!/usr/local/bin/python
import psp

#--Config---------------------------------------------------------

# Absolute path where the XML questionnaire and the product DB
# are located
base_dir = "/home/arno/public_html/corba/matrix/qtool/"

# URL to this script
script_url = "/cgi-bin/qtool.cgi"

# Do a backup of the database before each write operation
# A complete copy of the database will be made
do_backups = 1

# Send email.  Emails are sent to the maintainer when
# an error occurs or a new/updated profile has been
# submitted. If you say "1" here, you need to provide
# meaningful values for 'admin' and 'smtp_host' below.
send_email = 1

# admin email
admin = "arno@mico.org"

# SMTP host used for sending emails
smtp_host = "mico.org"

# known profiles
known_profiles = ['CORBA', 'DEMO']

# URLs of main pages
base_urls = {'CORBA': '/corba/matrix/',
             'DEMO'  : '/qtool/demo/'}

# Background colors used for product matrix
color_no      = "#CC4444"
color_yes     = "#44CC44"
color_unknown = "#CCCCCC"
color_planned = "#FFFF6B"


# Don't touch this one
arnos_layout = 0

#-------------------------------------------------------------------
#-------------------------------------------------------------------
#--- NO MORE CONFIGURATION BEYOND THIS POINT -----------------------
#-------------------------------------------------------------------
#-------------------------------------------------------------------

import cgi, pickle, os, sys, urllib, fcntl, whrandom, time
from xml.dom.minidom import parse, parseString
from stat import *
from FCNTL import LOCK_UN, LOCK_EX, LOCK_SH


#--Global parameters------------------------------------------------

# CGI parameters
params = cgi.SvFormContentDict()

# Profile name
profile = None

# DOM representation of questionnaire
profile_dom = None

# DB
db = None


#--Security stuff---------------------------------------------------

def send_admin_email (subject, msg):
    import smtplib

    if not send_email:
        return
    
    msg_hdr = ("From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n"
               % (admin, admin, subject))
    msg = msg_hdr + msg

    server = smtplib.SMTP(smtp_host)
    server.sendmail(admin, admin, msg)
    server.quit()


def security_alert (msg):
    # add Form parameters
    form = cgi.SvFormContentDict()
    msg = msg + "\r\n\r\nForm parameters:\r\n----------------\r\n"
    for field in form.keys():
        msg = msg + field + ": " + form[field] + "\r\n"

    # add environment parameters
    msg = msg + "\r\nEnvironment:\r\n------------\r\n"
    for field in os.environ.keys():
        msg = msg + field + ": " + os.environ[field] + "\r\n"

    send_admin_email ("CGI: SECURITY ALERT", msg)

    # generate message to submitter

    psp.emit_html ("Content-Type: text/html")
    psp.emit_html ("")
    psp.emit_html ("<HTML>")
    psp.emit_html ("<BODY>")
    psp.emit_html ("<H1>Boo!</H1>")
    psp.emit_html ("<P>")
    psp.emit_html ("Error!")
    psp.emit_html ("</BODY></HTML>")

    sys.exit()


#--DB stuff-------------------------------------------------------

def create_lock_file():
    lck_fn = base_dir + profile + ".db.lck"
    if not os.access (lck_fn, os.F_OK):
        # Lock file doesn't exist. Create it
        f = open (lck_fn, "w")
        f.close()
        
# acquire shared lock
def acquire_shared_lock():
    global lock
    create_lock_file()
    lock = open (base_dir + profile + ".db.lck", "r")
    fcntl.flock (lock.fileno(), LOCK_SH)

# acquire exclusive lock
def acquire_exclusive_lock():
    global lock
    create_lock_file()
    lock = open (base_dir + profile + ".db.lck", "r")
    fcntl.flock (lock.fileno(), LOCK_EX)

# release lock
def release_lock():
    global lock
    fcntl.flock (lock.fileno(), LOCK_UN)
    lock.close()

# Load DB
def load_db():
    global db
    try:
        acquire_shared_lock()
        db_fn = base_dir + profile + ".db"
        f = open (db_fn, "r")
        db = pickle.load (f)
        f.close()
        release_lock()
    except:
        db = {}
        release_lock()

# Write DB
def save_db():
    global db
    acquire_exclusive_lock()
    db_filename = base_dir + profile + ".db"
    if do_backups and os.access (db_filename, os.F_OK):
        # Do a backup of the old database
        db_bak = base_dir + profile
        db_bak += time.strftime("-%a-%d-%b-%Y-%H:%M:%S-%Z", time.localtime(time.time()))
        db_bak += ".db"
        os.rename (db_filename, db_bak)
    f = open (db_filename, "w")
    pickle.dump (db, f)
    f.close()
    release_lock()
    
#--Helper functions---------------------------------------------

def get_param (key):
    if params.has_key (key):
        return params[key]
    else:
        return ""


def cmp_fct (a, b):
    (a1, a2) = a
    (b1, b2) = b
    if a1 == b1:
        return 0
    if a1 < b1:
        return -1
    return 1

def sorted_product_list():
    l = []
    for i in db.keys():
        product = db[i]
        name = product.get ('Name', 'Noname')
        l.append ((name, i))
    l.sort (cmp_fct)
    pl = []
    for j in l:
        (a, b) = j
        pl.append (b)
    return pl

    
#--HTML prolog & epilog functions---------------------------------

javascript = """
<SCRIPT LANGUAGE="JavaScript">
   function popUp (URL)
   {
      doc = window.open(URL, 'Documentation', 'toolbar=0,scrollbars=1,location=0,statusbar=0,menubar=0,resizable=0,width=300,height=300');
   }
</SCRIPT>
"""

def inline_file_content (fn):
    f = open (base_dir + fn, "r")
    lines = f.readlines()
    for l in lines:

        psp.emit_html ("" + str ( l ) + "")

    f.close()


def write_arnos_html_prolog (title):

    psp.emit_html ("Content-Type: text/html")
    psp.emit_html ("")
    psp.emit_html ("<HTML>")
    psp.emit_html ("<HEAD>")
    psp.emit_html ("  <meta http-equiv=\"Content-Type\" content=\"text/html\">")
    psp.emit_html ("  <link rel=StyleSheet href=\"/css/zeld.css\" ")
    psp.emit_html ("        type=\"text/css\" media=\"screen\">")
    psp.emit_html ("  <link rel=StyleSheet href=\"/css/colors.css\" ")
    psp.emit_html ("        type=\"text/css\" media=\"screen\">")
    psp.emit_html ("  <TITLE>" + str ( title ) + "</TITLE>")
    psp.emit_html ("  " + str ( javascript ) + "")
    psp.emit_html ("</HEAD>")
    psp.emit_html ("<BODY>")
    psp.emit_html ("<br><a name=\"top\">&nbsp; </a><br>")
    psp.emit_html ("")
    psp.emit_html ("<table align=\"center\" width=\"92%\" border=\"0\" cellspacing=\"0\" cellpadding=\"1\">")
    psp.emit_html ("<tbody>")
    psp.emit_html ("<tr>")
    psp.emit_html ("  <td bgcolor=\"#333366\">")
    psp.emit_html ("  <table align=\"center\" width=\"100%\" border=\"0\" cellspacing=\"2\" cellpadding=\"20\">")
    psp.emit_html ("  <tbody>")
    psp.emit_html ("  <tr>")
    psp.emit_html ("  <td bgcolor=\"#ffffff\" width=\"100%\" valign=\"top\">")
    psp.emit_html ("")
    psp.emit_html ("   <p>")

    inline_file_content ("NAVBAR_HIERARCHY")
    inline_file_content ("NAVBAR_SET_main")

    
def write_plain_html_prolog (title):

    psp.emit_html ("Content-Type: text/html")
    psp.emit_html ("")
    psp.emit_html ("<HTML>")
    psp.emit_html ("<HEAD>")
    psp.emit_html ("  <TITLE>" + str ( title ) + "</TITLE>")
    psp.emit_html ("  " + str ( javascript ) + "")
    psp.emit_html ("</HEAD>")
    psp.emit_html ("<BODY>")


def write_html_prolog (title):
    if arnos_layout:
        write_arnos_html_prolog (title)
    else:
        write_plain_html_prolog (title)

def write_arnos_html_epilog():

    psp.emit_html ("</td>")
    psp.emit_html ("</tr>")
    psp.emit_html ("")
    psp.emit_html ("</tbody>")
    psp.emit_html ("</table>")

    psp.emit_html ("</td>")
    psp.emit_html ("</tr>")
    psp.emit_html ("</tbody>")
    psp.emit_html ("</table>")
    psp.emit_html ("<br><br>")
    psp.emit_html ("<br><br>")
    psp.emit_html ("</BODY>")
    psp.emit_html ("</HTML>")


def write_plain_html_epilog():

    psp.emit_html ("</BODY>")
    psp.emit_html ("</HTML>")


def write_html_epilog():
    if arnos_layout:
        write_arnos_html_epilog()
    else:
        write_plain_html_epilog()


#--Load questionnaire--------------------------------------------

def load_questionnaire():
    global profile_dom
    profile = get_param ("PROFILE")
    if profile not in known_profiles:
        security_alert ('Unknown profile: "' + profile + '"')

    dom_fn = base_dir + profile + ".xml"
    profile_dom = parse(dom_fn)
    # We could do thing a bit more efficiently here. Right now
    # the XML is parsed with _every_ invocation of this CGI
    # script. We could write out a pickled version that would
    # make things more efficient
    

#--DOM helper functions--------------------------------------------

def get_attribute (node, name):
    assert node.nodeType == node.ELEMENT_NODE
    attrs = node.attributes
    for i in range (0, attrs.length):
        a = attrs.item (i)
        assert a.nodeType == node.ATTRIBUTE_NODE
        if a.nodeName == name:
            return a.nodeValue
    return ""

def get_text (node):
    txt = ""
    for n in node.childNodes:
        if n.nodeType == node.TEXT_NODE:
            txt = txt + n.data
    return txt.strip()



#--Write a form----------------------------------------------------

# Write a query
def gen_input_field (name, type, editable, product_id):
    val = ""
    if product_id != None:
        val = db[product_id].get (name, "").strip()

    if not editable:
        val = cgi.escape (val)
        val = val.replace ("\012", "<br>")
        
    if type == "string":
        if editable:
            if product_id == None:

                psp.emit_html ("<input NAME=\"" + str ( name ) + "\" SIZE=25 MAXLENGTH=40>")

            else:

                psp.emit_html ("<input NAME=\"" + str ( name ) + "\" SIZE=25 MAXLENGTH=40 VALUE=\"" + str ( val ) + "\">")

        else:
            if val == "":
                val = "-"

            psp.emit_html ("<em>" + str ( val ) + "</em>")

            
    elif type == "textfield":
        if editable:

            psp.emit_html ("<textarea NAME=\"" + str ( name ) + "\" ROWS=6 COLS=30 WRAP=PHYSICAL>" + str ( val ) + "</textarea>")

        else:
            if val == "":
                val = "-"

            psp.emit_html ("<em>" + str ( val ) + "</em>")

            
    elif type == "feature":
        if editable:
            unknown_selected = ""
            yes_selected = ""
            no_selected = ""
            planned_selected = ""
            if val == "Unknown" or val == "":
                unknown_selected = " SELECTED"
            if val == "Yes":
                yes_selected = " SELECTED"
            if val == "No":
                no_selected = " SELECTED"
            if val == "Planned":
                planned_selected = " SELECTED"

            psp.emit_html ("<select NAME=\"" + str ( name ) + "\">")
            psp.emit_html ("<option VALUE=\"Unknown\"" + str ( unknown_selected ) + ">Unknown")
            psp.emit_html ("<option VALUE=\"Yes\"" + str ( yes_selected ) + ">Yes")
            psp.emit_html ("<option VALUE=\"No\"" + str ( no_selected ) + ">No")
            psp.emit_html ("<option VALUE=\"Planned\"" + str ( planned_selected ) + ">Planned")
            psp.emit_html ("</select>")

        else:

            psp.emit_html ("<em>" + str ( val ) + "</em>")

    else:
        assert 0


#   editable = true, product_id == None  : Empty form
#   editable = true, product_id != None  : Edit existing product
#   editable = false, product_id != None : Show product (read-only)
#   editable = false, product_id == None : Illegal

def write_form (editable, product_id = None):
    assert editable or product_id != None
    sections = profile_dom.getElementsByTagName ("section")
    for sec in sections:
        name = get_attribute (sec, "name")
        if editable:
            doc = get_text (sec.getElementsByTagName ("doc")[0])
        else:
            doc = ""

        psp.emit_html ("<hr>")
        psp.emit_html ("<h2>" + str ( name ) + "</h2>")
        psp.emit_html ("" + str ( doc ) + "")
        psp.emit_html ("<p>")
        psp.emit_html ("<table>")

        questions = sec.getElementsByTagName ("question")
        for q in questions:
            name = get_attribute (q, "name")
            type = get_attribute (q, "type")
            doc_url = script_url + '?PROFILE=' + profile
            doc_url += '&ACTION=DOC&NAME=' + name
            query = get_text (q.getElementsByTagName ("query")[0])

            psp.emit_html ("<tr>")
            psp.emit_html ("  <td valign=\"top\">")
            psp.emit_html ("    <a href=\"javascript:popUp ('" + str ( doc_url ) + "')\" tabindex=\"42\">" + str ( query ) + "</a>:")
            psp.emit_html ("  </td>")
            psp.emit_html ("  <td>")
            psp.emit_html ("    &nbsp;")
            psp.emit_html ("  </td>")
            psp.emit_html ("  <td>")

            gen_input_field (name, type, editable, product_id)

            psp.emit_html ("  </td>")
            psp.emit_html ("</tr>")


        psp.emit_html ("</table>")

    

#--New profile----------------------------------------------------

def action_new():
    write_html_prolog ('New ' + profile + ' Product Profile')

    psp.emit_html ("<h1>New " + str ( profile ) + " Product Profile</h1>")
    psp.emit_html ("<form METHOD=\"POST\" ACTION=\"" + str ( script_url ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"PROFILE\" VALUE=\"" + str ( profile ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"ACTION\" VALUE=\"SUBMIT\">")
    psp.emit_html ("<input TYPE=\"SUBMIT\" VALUE=\"Submit questionnaire\">")
    psp.emit_html ("&nbsp;")
    psp.emit_html ("<INPUT TYPE=\"SUBMIT\" NAME=\"CANCEL\" VALUE=\"Cancel\">")

    # Write an empty, editable form
    write_form (1)

    psp.emit_html ("<P>&nbsp;<P>")
    psp.emit_html ("<input TYPE=\"SUBMIT\" VALUE=\"Submit questionnaire\">")
    psp.emit_html ("&nbsp;")
    psp.emit_html ("<INPUT TYPE=\"SUBMIT\" NAME=\"CANCEL\" VALUE=\"Cancel\">")
    psp.emit_html ("</form>")

    write_html_epilog()
    

#--Edit profile----------------------------------------------------

def action_edit():
    product_id = get_param ("PRODUCT_ID")
    write_html_prolog ('Edit ' + profile + ' Product Profile')

    psp.emit_html ("<h1>Edit " + str ( profile ) + " Product Profile</h1>")
    psp.emit_html ("<form METHOD=\"POST\" ACTION=\"" + str ( script_url ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"PROFILE\" VALUE=\"" + str ( profile ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"ACTION\" VALUE=\"SUBMIT\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"PRODUCT_ID\" VALUE=\"" + str ( product_id ) + "\">")
    psp.emit_html ("<input TYPE=\"SUBMIT\" VALUE=\"Submit questionnaire\">")
    psp.emit_html ("&nbsp;")
    psp.emit_html ("<INPUT TYPE=\"SUBMIT\" NAME=\"CANCEL\" VALUE=\"Cancel\">")

    # Write an editable form
    write_form (1, product_id)

    psp.emit_html ("<P>&nbsp;<P>")
    psp.emit_html ("<input TYPE=\"SUBMIT\" VALUE=\"Submit questionnaire\">")
    psp.emit_html ("&nbsp;")
    psp.emit_html ("<INPUT TYPE=\"SUBMIT\" NAME=\"CANCEL\" VALUE=\"Cancel\">")
    psp.emit_html ("</form>")

    write_html_epilog()
    

#--Show--------------------------------------------------------

def action_show():
    write_html_prolog (profile + ' Product Profile')
    product_id = get_param ("PRODUCT_ID")

    psp.emit_html ("<h1>" + str ( profile ) + " Product Profile</h1>")
    psp.emit_html ("Last updated: <em>" + str ( db[product_id].get ('LAST_UPDATED', '-') ) + "</em>")
    psp.emit_html ("<p>")
    psp.emit_html ("<form METHOD=\"POST\" ACTION=\"" + str ( script_url ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"PROFILE\" VALUE=\"" + str ( profile ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"PRODUCT_ID\" VALUE=\"" + str ( product_id ) + "\">")
    psp.emit_html ("<input TYPE=\"HIDDEN\" NAME=\"ACTION\" VALUE=\"EDIT\">")
    psp.emit_html ("<input TYPE=\"SUBMIT\" VALUE=\"Edit this profile\">")
    psp.emit_html ("</form>")

    # Write an empty, non-editable form
    write_form (0, product_id)
    write_html_epilog()
    

#--Index-------------------------------------------------------

def action_index():

    psp.emit_html ("Content-Type: text/html")
    psp.emit_html ("")
    psp.emit_html ("<UL>")

    for i in sorted_product_list():
        product = db[i]
        name = product.get ('Name', '')
        if name == "":
            name = 'Noname product'
        vendor = product.get ('Vendor', '')
        if vendor != "":
            vendor = ' (' + vendor + ')';

        url_prefix = script_url + "?PROFILE=" + profile
        url_prefix += "&PRODUCT_ID=" + i
        url_prefix += "&ACTION="
        

        psp.emit_html ("<LI> <a href=\"" + str ( url_prefix + 'SHOW' ) + "\">" + str ( name ) + "" + str ( vendor ) + "</a>")
        psp.emit_html ("&nbsp;<font size=\"-1\">(<a href=\"" + str ( url_prefix + 'EDIT' ) + "\">Edit</a>)</font>")

        

    psp.emit_html ("</UL>")

    

#--Submit product-------------------------------------------------

def action_submit():
    product_id = get_param ("PRODUCT_ID")
    if product_id == "":
        # This is a new product. Generate unique ID
        product_id = str (whrandom.randint (0, 1000000))
        while db.has_key (product_id):
            product_id = str (whrandom.randint (0, 1000000))
    questions = profile_dom.getElementsByTagName ("question")
    product = {}
    product['LAST_UPDATED'] = time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.localtime(time.time()))
    for q in questions:
        name = get_attribute (q, "name")
        product[name] = get_param (name)
    db[product_id] = product
    save_db()

    # Send email notification to maintainer
    msg = "User from " + os.environ["REMOTE_ADDR"]
    msg += " submitted profile " + product_id
    send_admin_email ("QTool: " + profile + " profile update", msg)
    
    write_html_prolog("Thank you")

    psp.emit_html ("<h1>" + str ( profile ) + " Product Profile</h1>")
    psp.emit_html ("<h2>Thank you</h2>")
    psp.emit_html ("Thank you for submitting a " + str ( profile ) + " product profile.")
    psp.emit_html ("The database has been updated. Click")
    psp.emit_html ("<a href=\"" + str ( base_urls[profile] ) + "\">here</a> to return to")
    psp.emit_html ("the main page.")

    write_html_epilog()


#--Documentation--------------------------------------------------

def action_doc():
    write_plain_html_prolog ('Documentation')
    name = get_param ("NAME")
    questions = profile_dom.getElementsByTagName ("question")
    for q in questions:
        if name == get_attribute (q, "name"):
            doc = get_text (q.getElementsByTagName ("doc")[0])
            query = get_text (q.getElementsByTagName ("query")[0])

            psp.emit_html ("<h1>" + str ( query ) + "</h1>")
            psp.emit_html ("" + str ( doc ) + "")
            psp.emit_html ("<p>")
            psp.emit_html ("<form>")
            psp.emit_html ("<input type=button value=\"Close Window\"")
            psp.emit_html ("       onClick=\"javascript:window.close();\">")
            psp.emit_html ("</form>")

    write_plain_html_epilog()

    
#--Matrix--------------------------------------------------------

def get_section (question_name):
    questions = profile_dom.getElementsByTagName ("question")
    for i in questions:
        name = get_attribute (i, "name")
        if name == question_name:
            # We found the right question. Now look in which
            # section it is defined
            section = i.parentNode
            return get_attribute (section, "name")
    assert 0


def write_matrix_header (cols):

    psp.emit_html ("<TR>")
    psp.emit_html ("<TH ROWSPAN=\"2\">Product</TH>")


    # Print the row with the general sections
    current_section = ""
    col_names = ""
    num = 0
    cols_and_x = cols + ['X']
    
    for i in cols_and_x:
        # Get name of section of the current column
        if i != 'X':
            section = get_section (i)
        else:
            section = '__X__'

        if current_section == "":
            current_section = section
            col_names = "|" + i + "|"
            num = 1
            continue

        if section == current_section:
            # the idea of col_names is that we generate a string
            # that has roughly the length of that section of the
            # table. We use it as a reference to see if the section
            # heading will fit or not
            col_names += "|" + i + "|"
            num += 1
            continue

        section_name = current_section
        # if the section name is too long, we just leave it empty
        if len (section_name) > len (col_names):
            section_name = "&nbsp;"

        psp.emit_html ("<TH COLSPAN=\"" + str ( num ) + "\"><CENTER>" + str ( section_name ) + "</CENTER></TH>")

        current_section = section
        col_names = "|" + i + "|"
        num = 1

    psp.emit_html ("</TR>")
    psp.emit_html ("<TR>")


    # now write out all the columns
    for i in cols:

        psp.emit_html ("<TH>" + str ( i ) + "</TH>")


    psp.emit_html ("</TR>")


def write_yes():

    psp.emit_html ("<TD BGCOLOR=\"" + str ( color_yes ) + "\"><CENTER>Y</CENTER></TD>")

    
def write_no():

    psp.emit_html ("<TD BGCOLOR=\"" + str ( color_no ) + "\"><CENTER>-</CENTER></TD>")

    
def write_planned():

    psp.emit_html ("<TD BGCOLOR=\"" + str ( color_planned ) + "\"><CENTER>P</CENTER></TD>")

    
def write_unknown():

    psp.emit_html ("<TD BGCOLOR=\"" + str ( color_unknown ) + "\"><CENTER>?</CENTER></TD>")


def write_feature (feature):
    if feature == "Yes":
        write_yes()
    if feature == "No":
        write_no()
    if feature == "Planned":
        write_planned()
    if feature == "Unknown":
        write_unknown()


def action_matrix():

    psp.emit_html ("Content-Type: text/html")
    psp.emit_html ("")
    psp.emit_html ("<TABLE ID=matrix CELLPADDING=2 CELLSPACING=2 BORDER=1 BGCOLOR=\"#d0d0d0\">")

    cols = get_param ("COLS").split ('_')

    write_matrix_header (cols)

    # Iterate over all products
    for i in sorted_product_list():

        psp.emit_html ("<TR>")

    
        product = db[i]
        name = product.get ('Name', '')
        if name == "":
            name = 'Noname product'
        url = script_url + "?PROFILE=" + profile
        url += "&PRODUCT_ID=" + i
        url += "&ACTION=SHOW"

        psp.emit_html ("<TD><a href=\"" + str ( url ) + "\">" + str ( name ) + "</a></TD>")

        for j in cols:
            write_feature (product.get (j, "Unknown"))


        psp.emit_html ("</TR>")


    psp.emit_html ("</TABLE>")


#--Shell List-----------------------------------------------------

def shell_list():
    for i in db.keys():
        product = db[i]
        name = product.get ("Name", "-")
        print 'Profile', i, ':', name

#--MAIN-----------------------------------------------------------

def redirect():

    psp.emit_html ("Content-Type: text/html")
    psp.emit_html ("")
    psp.emit_html ("<html>")
    psp.emit_html ("<head>")
    psp.emit_html ("<meta HTTP-EQUIV=\"REFRESH\" CONTENT=\"0;" + str ( base_urls[profile] ) + "\">")
    psp.emit_html ("</head>")
    psp.emit_html ("")
    psp.emit_html ("<body>")
    psp.emit_html ("</body>")
    psp.emit_html ("</html>")



# This function gets called for CGI scripts
def main():
    global profile
    profile = get_param ("PROFILE")
    
    # first check if the Form has a cancel button
    if params.has_key ("CANCEL"):
        # Yes, simply redirect to main page
        redirect()
        return

    # Before we do the computation intensive operations
    # (reading database, XML specs for questionnaire, etc)
    # first activate the cache to see if we can take a shortcut

    # Key used for cache. Query string is just perfect for this
    cache_key = os.environ['QUERY_STRING']
    # Timestamp of DB. If the cache is older, we need to
    # rebuild the HTML output
    db_timestamp = None
    try:
        db_fn = base_dir + profile + ".db"
        db_timestamp = os.stat (db_fn)[ST_MTIME]
    except:
        pass


    psp.cache_directive ( cache_key, db_timestamp )

    
    action = get_param ("ACTION")
    load_questionnaire()
    load_db()
    
    if action == "NEW":
        action_new ()
    elif action == "DOC":
        action_doc ()
    elif action == "SUBMIT":
        action_submit()
    elif action == "EDIT":
        action_edit()
    elif action == "INDEX":
        action_index()
    elif action == "SHOW":
        action_show()
    elif action == "MATRIX":
        action_matrix()
    else:
        security_alert ('Bad action "' + action + '"')


# We have been called from CGI script
try:
    main()
except SystemExit:
    # we exited from PSP because of a cache hit. Do nothing
    pass
except:
    from traceback import *
    (ex_type, ex_value, ex_traceback_obj) = sys.exc_info()
    ex_traceback_list = format_list (extract_tb (ex_traceback_obj))
    ex_traceback = ""
    for i in ex_traceback_list:
        ex_traceback += i
    security_alert ('Exception: %s : %s\r\n%s' %
                    (ex_type, ex_value, ex_traceback))
                    
sys.exit()
